part of fields.model;

class Field<T> {
    
  Function _getter;
  Function _setter;
  
  final Iterable<FieldAnnotation> annotations;
  
  FieldList _children;
  /// The children of this field.
  FieldList get children => _children;
  
  /// A description of this field; this is `null` by default.
  final String description;
  
  /// The display name of this field; this is equal to the `name` of this field
  /// by default.
  final String displayName;
  
  /// The type of this field's editor; this is `null` by default.
  final Type editorType;
  
  /// Returns `true` if this field is private.
  final bool isPrivate;
  
  /// Returns `true` if this field is read-only.
  final bool isReadOnly;
  
  /// The name of this field.
  final String name;

  /// The parent of this field or `null` if this field has no parent.
  final Field parent;
      
  /// This field's preferred transform; this is `null` by default.
  final ValueTransform transform;
  
  /// The type of this field's [value].
  final Type type;
  
  Object _value;
  /// The current value of this field.
  Object 
    get value => _value;
    set value(Object value) {
      if(isReadOnly) {
        throw new StateError('Field $name is read-only.');
      }
      if (value != _value) {
        _setter(value);
        _value = value;
        _refreshChildren().then((_) {
          _values.add(value);
          if (parent != null) {
            parent._values.add(parent._value);
          }
        });
      }
    }
  
  final _values = new StreamController<Object>.broadcast();
  /// An observable stream of this field's values as they change.
  Stream<Object> get values => _values.stream;
      
  Field(
      String name,
      T getter(),
     {setter(T),
      Type type,
      this.parent,
      this.editorType,
      this.transform,
      this.annotations: const <FieldAnnotation>[],
      this.description,
      String displayName,
      this.isPrivate: false,
      bool isReadOnly})      
      : this.name = name
      , this.displayName = (displayName == null) ? name : displayName
      , this._getter = getter
      , this._setter = setter
      , this.type = (type == null) ? T : type
      , this.isReadOnly = (isReadOnly == null) ? (setter == null) : isReadOnly {
    assert(name != null);
    assert(getter != null);
    _children = new FieldList(this);
    _value = _getter();
    _children.refreshSource(_value);
  }
  
  /// Refresh this field's [value].
  /// 
  /// A field can only know about changes to its [value] that happen through
  /// this interface.  If another object changes the underlying value in some 
  /// other way then this method should be called.
  void refreshValue() {
    final value = _getter();
    if (value != _value) {
      _value = value;
      _refreshChildren().then((_) => _values.add(value));
    }
  }
        
  Future _refreshChildren() {    
    if (children.isEmpty) {
      return new Future.value();
    }
    children.refreshSource(_value);
    return Future.forEach(children, (f) => new Future(f.refreshValue));
  }
}
